/**
 * Copyright (c) 2016-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#import <React/RCTUtils.h>

#import "WebAnalyticsModule.h"

static NSString *const RNFBErrorDomain = @"RNFBErrorDomain";

@implementation WebAnalytics {
    BOOL _debugMode;
    WKWebView *_webView;
    NSString *_userId;
    RCTPromiseResolveBlock _webviewInitResolve;
    RCTPromiseRejectBlock _webviewInitReject;
}
#pragma mark -
#pragma mark Module Setup

  RCT_EXPORT_MODULE();

  + (BOOL)requiresMainQueueSetup
  {
      return YES;
  }

  - (instancetype)init
  {
      if (self = [super init]) {
          _debugMode = NO;
          _webView = nil;
          _userId = nil;
      }
      return self;
  }

  - (dispatch_queue_t)methodQueue {
    return dispatch_get_main_queue();
  }

#pragma mark -
#pragma mark Firebase Analytics Methods

  RCT_EXPORT_METHOD(debugMode:(BOOL) enable
                    resolver:
                    (RCTPromiseResolveBlock) resolve
                    rejecter:
                    (RCTPromiseRejectBlock) reject) {
      _debugMode = enable;
      return resolve([NSNull null]);
  }

    RCT_EXPORT_METHOD(initializeWebMode:(NSString *) rawHTML
                  resolver:
                  (RCTPromiseResolveBlock) resolve
                  rejecter:
                  (RCTPromiseRejectBlock) reject) {
        
        if (_webviewInitResolve || _webviewInitReject) {
            //unlock pending promise
            _webviewInitReject(@"Error", @"Kill pending promise", nil);
        }
        
        if (_webView != nil) {
            [_webView removeFromSuperview];
            _webView = nil;
            [self removeFromSuperview];
        }
        
        WKWebViewConfiguration *wkWebViewConfig = [WKWebViewConfiguration new];
        WKPreferences *prefs = [[WKPreferences alloc]init];
        [wkWebViewConfig setValue:@TRUE forKey:@"allowUniversalAccessFromFileURLs"];
        [prefs setValue:@TRUE forKey:@"allowFileAccessFromFileURLs"];
        wkWebViewConfig.preferences = prefs;
        wkWebViewConfig.websiteDataStore = [WKWebsiteDataStore defaultDataStore];
        
        _webView = [[WKWebView alloc] initWithFrame:CGRectZero configuration:wkWebViewConfig];
        _webView.navigationDelegate = self;
        
        [self addSubview:_webView];
        
        UIViewController *rootController = RCTPresentedViewController();
        [rootController.view addSubview:self];
        
        [_webView loadHTMLString:rawHTML baseURL:[NSURL URLWithString:@"http://localhost"]];
#if defined(__IPHONE_OS_VERSION_MAX_ALLOWED) && __IPHONE_OS_VERSION_MAX_ALLOWED >= 160400
        if (@available(iOS 16.4, *)) {
            [_webView setInspectable:_debugMode];
        } else {
            // Fallback on earlier versions
        }
#endif
            
        _webviewInitResolve = resolve;
        _webviewInitReject = reject;
    }

  RCT_EXPORT_METHOD(logEvent:
    (NSString *) name
        params:
        (NSDictionary *) params
        resolver:
        (RCTPromiseResolveBlock) resolve
        rejecter:
        (RCTPromiseRejectBlock) reject) {
    @try {
        
        [self logEventToWebView:name andData:params];
    } @catch (NSException *exception) {
      [self rejectPromiseWithExceptionDict:reject exception:exception];
    }

    resolve([NSNull null]);
  }

  RCT_EXPORT_METHOD(logPageView:
        (NSDictionary *) params
        resolver:
        (RCTPromiseResolveBlock) resolve
        rejecter:
        (RCTPromiseRejectBlock) reject) {
    @try {
      
        [self logPageViewToWebView:params];
    } @catch (NSException *exception) {
      [self rejectPromiseWithExceptionDict:reject exception:exception];
    }

    resolve([NSNull null]);
  }

  RCT_EXPORT_METHOD(logScreenView:
      (NSDictionary *) params
      resolver:
      (RCTPromiseResolveBlock) resolve
      rejecter:
      (RCTPromiseRejectBlock) reject) {
    @try {
    
        [self logScreenViewToWebView:params];
    } @catch (NSException *exception) {
      [self rejectPromiseWithExceptionDict:reject exception:exception];
    }

    resolve([NSNull null]);
  }

  RCT_EXPORT_METHOD(setUserId:
    (NSString *) id
        resolver:
        (RCTPromiseResolveBlock) resolve
        rejecter:
        (RCTPromiseRejectBlock) reject) {
    @try {
        _userId = id;
    } @catch (NSException *exception) {
        [self rejectPromiseWithExceptionDict:reject exception:exception];
    }
    resolve([NSNull null]);
  }

RCT_EXPORT_METHOD(addInstance:(NSDictionary *)config
                  resolver:(RCTPromiseResolveBlock) resolve
                  rejecter:(RCTPromiseRejectBlock) reject) {
    NSString *name = [config objectForKey:@"name"];
    [self addInstanceToWebView:name withConfig:config];
    
    resolve([NSNull null]);
}

RCT_EXPORT_METHOD(removeInstance:(NSString *)name
                  resolver:(RCTPromiseResolveBlock) resolve
                  rejecter:(RCTPromiseRejectBlock) reject) {
    [self removeInstanceFromWebView:name];
    
    resolve([NSNull null]);
}

#pragma mark -
#pragma mark Private methods

- (void)evaluateJavaScript:(NSString *)js {
    if (_webView != nil) {
        [_webView evaluateJavaScript:js completionHandler:nil];
    }
}

- (NSString *)getJSONParams:(NSDictionary *)data {
    NSError *error;
    NSDictionary *eventParams = [self cleanJavascriptParams:data];
    NSData *jsonData = [NSJSONSerialization dataWithJSONObject:eventParams options:0 error:&error];
    if (!jsonData) {
        NSLog(@"Got an error: %@", error);
        return @"{}";
    }
    
    return [[NSString alloc] initWithData:jsonData encoding:NSUTF8StringEncoding];
}

- (void)logEventToWebView:(NSString *)name andData:(NSDictionary *)data {
    [self evaluateJavaScript:[NSString stringWithFormat:@"logEvent('%@', '%@')", name, [self getJSONParams:data]]];
}

- (void)logPageViewToWebView:(NSDictionary *)data {
    [self evaluateJavaScript:[NSString stringWithFormat:@"logPageView('%@')", [self getJSONParams:data]]];
}

- (void)logScreenViewToWebView:(NSDictionary *)data {
    [self evaluateJavaScript:[NSString stringWithFormat:@"logScreenView('%@')", [self getJSONParams:data]]];
}

- (void)addInstanceToWebView:(NSString *)name withConfig:(NSDictionary *)config {
    if (_webView != nil) {
        NSError *error;
        
        NSMutableDictionary *newConfig = [config mutableCopy];
        [newConfig setValue:@(_debugMode) forKey:@"debug_mode"];
        if (_userId != nil) {
            [newConfig setValue:_userId forKey:@"user_id"];
        }

        NSData *jsonConfig = [NSJSONSerialization dataWithJSONObject:newConfig options:0 error:&error];
        if (!jsonConfig) {
            NSLog(@"Got an error: %@", error);
        } else {
            NSString *jsonString = [[NSString alloc] initWithData:jsonConfig encoding:NSUTF8StringEncoding];
            [_webView evaluateJavaScript:[NSString stringWithFormat:@"addProject('%@', '%@')", name, jsonString] completionHandler:nil];
        }
    }
}

- (void)removeInstanceFromWebView:(NSString *)name{
    if (_webView != nil) {
        [_webView evaluateJavaScript:[NSString stringWithFormat:@"deleteProject('%@')", name] completionHandler:nil];
    }
}

  - (NSDictionary *)cleanJavascriptParams:(NSDictionary *)params {
    NSMutableDictionary *newParams = [params mutableCopy];
    if (newParams[@"items"]) {
      NSMutableArray *newItems = [NSMutableArray array];
      [(NSArray *)newParams[@"items"] enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        NSMutableDictionary *item = [obj mutableCopy];
        if (item[@"quantity"]) {
          item[@"quantity"] = @([item[@"quantity"] integerValue]);
        }
        [newItems addObject:[item copy]];
      }];
      newParams[@"items"] = [newItems copy];
    }
    return [newParams copy];
  }

  - (void)rejectPromiseWithExceptionDict:(RCTPromiseRejectBlock)reject exception:(NSException *)exception {
    NSMutableDictionary *userInfo = [NSMutableDictionary dictionary];

    [userInfo setValue:@(YES) forKey:@"fatal"];
    [userInfo setValue:@"unknown" forKey:@"code"];
    [userInfo setValue:exception.reason forKey:@"message"];
    [userInfo setValue:exception.name forKey:@"nativeErrorCode"];
    [userInfo setValue:exception.reason forKey:@"nativeErrorMessage"];

    NSError *error = [NSError errorWithDomain:RNFBErrorDomain code:666 userInfo:userInfo];

    reject(exception.name, exception.reason, error);
  }

- (void) webView:(WKWebView *)webView didFinishNavigation:(WKNavigation *)navigation {
    if (_webviewInitResolve) {
        _webviewInitResolve([NSNull null]);
        
        _webviewInitResolve = nil;
        _webviewInitReject = nil;
    }
}

- (void) webView:(WKWebView *)webView didFailNavigation:(WKNavigation *)navigation withError:(NSError *)error {
    if (_webviewInitReject) {
        _webviewInitReject(@"Error", error.description, error);
        
        _webviewInitResolve = nil;
        _webviewInitReject = nil;
    }
}

@end
